/*!
\file mollweide.cpp
\brief Class definition for Mollweide
\author Nicholas Phillips, phillips@friday.gsfc.nasa.gov
\date Feb 23, 2004
*/
#include <iostream>
#include "mollweide.h"

using namespace std;

//=====================================================================================
double toMollweide(const double phi, const double lambda, double &x, double &y)
{
	x = phi;
	y = lambda;

	const double eps = 1e-6; 	// how well we solve for theta
	const int nmax = 20;		// max number of iterations for Newton's method
	const double lambda0 = 0;	// offset
	// Need to solve 2theta + sin(2theta) == pi sin(phi), for theta
	double theta;
	// Upper limit
	if( fabs(phi - M_PI/2) < 1e-6) {
		theta = M_PI/2;
	}
	// Lower limit
	else if(fabs(phi + M_PI/2) < 1e-6 ) {
		theta = -M_PI/2;
	}
	else {
		double dtheta;
		double f,df;
		// Use Newton's method to solve
		// Initial guess
		theta = phi/2;
		// function for whose zero we solve for
		f = 2*theta + sin(2*theta) - M_PI*sin(phi);
		int i;
		for(i = 0; i < nmax; i++) {
			// are we close enough to zero?
			if( fabs(f) < eps)
				break;
			// derivation of f
			df = 2*(1+cos(2*theta));
			// Newton's method error
			dtheta = -f/df;
			// update guess
			theta += dtheta;
			// new value of function
			f = 2*theta + sin(2*theta) - M_PI*sin(phi);
		}
	}
	// We now have the theta we need
	x = 2*sqrt(2.)*(lambda- lambda0)*cos(theta)/M_PI;
	y = sqrt(2.)*sin(theta);
	return theta;
}

//=====================================================================================
void Face::toMollweide()
{
	double r = 1/sqrt(2.);
	double phi;
	double lambda;
	double x,y;

	GLVertLI qli = quadList.begin();
	GLVertLI qli_end = quadList.end();
	GLVertI pti;
	GLVertI pti_end;
	for(; qli !=  qli_end; ++qli) {
		pti_end = qli->end();
		for(pti = qli->begin(); pti != pti_end; ++pti) {
			phi = asin(pti->z);
			lambda = atan2(pti->y,pti->x);
			::toMollweide(phi,  lambda, x, y);
			if( (face == 2 || face == 10) && x > 0) x *= -1;
			pti->y = r*x;
			pti->z = r*y;
			pti->x = 0;
		}
	}
	if( face == 6 )
		toMollweideBackfaceSplit();
	return;
}
//=====================================================================================
void Face::toMollweideBackfaceSplit()
{
	if( face != 6 )
		return;
	
	// copy the current list to temp storage
	GLVertList oldquads;
	oldquads.resize(quadList.size());
	GLVertLI srcqli = quadList.begin();
	GLVertLI destqli = oldquads.begin();
	for(uint i = 0; i < quadList.size(); i++) {
		destqli->resize(srcqli->size());
		GLVertI srcpti = srcqli->begin();
		GLVertI destpti = destqli->begin();
		for( uint j = 0; j < srcqli->size(); j++)
			*destpti++ = *srcpti++;
		srcqli++;
		destqli++;
	}
	
	// Now refill it, making the break along the back as needed
	srcqli = oldquads.begin();
	GLVertLI qli;
	quadList.resize(0);
	uint i = 0;
	for(srcqli = oldquads.begin(); srcqli != oldquads.end(); ++srcqli) {
		i++;
		
		bool patch = true;	// need to patch the end of the left and strart of the right sides
		bool leftside = true;	// true while building the left side quadlist
		
		GLVertices newverts1;
		GLVertices newverts2;
		
		uint jbreak = srcqli->size(); // vertex number of switch between left and right side
		jbreak = 2*i-2;
		for(uint j = 0; j < srcqli->size(); j++ ){
			
			GLPoint p = (*srcqli)[j];
			if( leftside ) {
				p.y = -fabs(p.y);
				newverts1.push_back(p);
				leftside = j < jbreak;
			}
			else if( patch ) {
				
				p.y = -fabs(p.y);
				newverts1.push_back(p);
				
				GLPoint p2;
				GLVertLI qli = srcqli;
				qli++;
				if( qli != oldquads.end() ) {
					p2 = (*qli)[j+1];
					p2.y = -fabs(p2.y);
				}
				else {
					p2.y = -1.66597;
					p2.z =  0.553293;
					p2.s = 0.75;
					p2.t = 0.5;
				}
				//cout << "\tp2.s:  " << setw(8) << p2.s << ", p2.t:  " << setw(8) << p2.t;
				newverts1.push_back(p2);
				newverts1.push_back(p);
				
				
				// Patch the beginning of the right side of the split face
				p = (*srcqli)[j+1];
				p.y = fabs(p.y);
				if( srcqli != oldquads.begin() ) {
					qli = srcqli;
					qli--;
					p2 = (*qli)[j];
					p2.y = fabs(p2.y);
				}
				else {
					p2.y =  1.66597;
					p2.z = -0.553293;
					p2.s = 0.5;
					p2.t = 0.25;
				}
				//cout << "\tp2.s:  " << setw(8) << p2.s << ", p2.t:  " << setw(8) << p2.t;
				newverts2.push_back(p);
				newverts2.push_back(p2);
				
				patch = false;
				//cout << "p";
			}
			else {
				p.y = fabs(p.y);
				newverts2.push_back(p);
			}
			
		}
		quadList.push_back(newverts1);
		quadList.push_back(newverts2);
	}
	return;
}
//=====================================================================================
void Mollweide::setResolution(int sphere_res_)
{
	Sphere3D::setResolution(sphere_res_);
	for(int i = 0; i < 12; i++)
		faces[i].toMollweide();

	return;
}

//=====================================================================================
bool Mollweide::select(const qglviewer::Vec &o,const qglviewer::Vec &d, qglviewer::Vec &v)
{
	return false;
}
