/*!
\file skyviewer.cpp
\brief Class definition for SkyViewer
\author Nicholas Phillips, phillips@friday.gsfc.nasa.gov
\date Feb 23, 2004
*/

#include <qimage.h>
#include <qfiledialog.h>
#include "skyviewer.h"
#include "colortable.h"
#include "timer.h"

using namespace qglviewer;
using namespace std;

extern QString incomingfname;

//====================================================================
// Constructor must call the base class constructor.
SkyViewer::SkyViewer(QWidget *parent,const char *name) : QGLViewer(parent, name)
{
	constraint = new qglviewer::CameraConstraint(camera());
	cwd = ".";

}

//====================================================================
SkyViewer::~SkyViewer()
{
	delete constraint;
}

//====================================================================
float SkyViewer::getmax() {
	switch( current_value ){
		case temperature: return max.I;
		case Qpolarization: return max.Q;
		case Upolarization: return max.U;
		case NumObs: return max.Nobs;
	}
	return 0.0;
}

//====================================================================
void SkyViewer::init()
{
	//setBackgroundColor(0,0,0);
	setBackgroundColor(QColor(0,0,0));
	//setBackgroundColor(1,1,1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//setDisplayFPS(true);
	setFPSIsDisplayed(true);
	GLfloat light_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat light_zero[] = { 0.0, 0.0, 0.0, 1.0 };
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_zero);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_zero);

	// Enable GL textures
	//glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	//glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glEnable( GL_TEXTURE_2D );

	// Nice texture coordinate interpolation
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	//camera()->setPosition(3.0,0,0);
	//camera()->setOrientation(M_PI/2,0);
	camera()->setPosition( Vec(3.0,0,0) );
	camera()->lookAt( Vec(0,0,0) );
	camera()->setUpVector(Vec(0,0,1));

	firstcall = true;
	current_mode = sphere; 
	current_value = temperature;
	tex_res = 128;
	sphere_res = 16;

	sphere3d_proj.setResolution(sphere_res);
	mollweide_proj.setResolution(sphere_res);

	projection = &sphere3d_proj;

	for(ncolors = 0; colortable[ncolors][0] != -1; ncolors++);
	colors.resize(ncolors);
	for(int i = 0; i < ncolors; i++)
		colors[i] = qRgb((int)(255*colortable[i][0]),(int)(255*colortable[i][1]),(int)(255*colortable[i][2]));

	//loadImage();
	//loadFile();
	startAnimation();

	return;
}


//====================================================================
void SkyViewer::draw()
{
	
	if( firstcall ) {
		firstcall = false;
		loadFile();
	}

	glDisable( GL_TEXTURE_2D );
	selector.draw();

	glColor4f(1,1,1,1);
	glEnable( GL_TEXTURE_2D );
	projection->draw();
	
	return;
}
//====================================================================
void SkyViewer::animate()
{
	selector.animate();
	return;
}

//====================================================================
void SkyViewer::saveSnapshot()
{
	QGLViewer::saveSnapshot(false);
	return;
}

//====================================================================
void SkyViewer::keyPressEvent(QKeyEvent *e)
{
	switch (e->key()) {
		default: QGLViewer::keyPressEvent(e);
	}
}

//====================================================================
QString SkyViewer::helpString() const
{
	QString text("<h2>S ky V i e w e r</h2>");
	text += "View a <b>Healpix</b>-based skymap, stored in a <b>FITS</b> file.<br><br>";
	text += "Use <b>WhatsThis</b> for help on each of the controls.<br>";
	text += "See the <b><u>K</u>eyboard</b> and <b><u>M</u>ouse</b> tabs above for help ";
	text += "on view controls for the main OpenGL viewing window.<br><br>";
	text += "To view available information for a pixel, use ";
	text += "<b>Shift</b>+<b>Left Mouse Click</b> to select pixel.<br>";
	text += "Presently, this only works in <b>3D Sphere</b> display mode.<br><br>";
	text += "Program: skyviewer<br>";
	text += "Version: 0.2.2<br>";
	text += "Date: Feb 22, 2005<br>";
	text += "Author: Nicholas Phillips<br>";
	text += "Email: phillips@friday.gsfc.nasa.gov<br>";
	text += "Website: http://lambda.gsfc.nasa.gov/toolbox/tb_skyviewer_ov.cfm<br>";

	return text;
}

//===============================================================
// My Slots
//===============================================================
//-------------------------------------------------
void SkyViewer::loadFile()
{
    //cout << "SkyViewer::loadFile()" << endl;
	bool need_to_load = true;
	/* Sit in this loop until a valid file is selected
	    or user cancels load request. */
	while( need_to_load ){
		currentfname = ! incomingfname.isEmpty() ? incomingfname :
			QFileDialog::getOpenFileName(cwd, "FITS (*.fits);;All Files (*)",
												this, "Choose", "Select a FITS skymap");
		// In case of Cancel
		if (currentfname.isEmpty())
			return;

		incomingfname = ""; // mark as not to be used any more
		//cout << "Selected: " << name << endl;
		if( ! sky.loadfitsfile(currentfname) )
			cerr << "SkyViewer::loadFile:failed to find valid fits file" << endl;
		else
			need_to_load = false;

		// set the current working directory to where we found this file
		// by using the whole filename, next dialog call pre-selects the file.
		cwd = currentfname;
	}

	max0.I = max0.Q = max0.U = max0.Nobs = 0;
	for(unsigned int i = 0; i < sky.npixels(); i ++) {
		if( fabs(sky[i].I) > max0.I    ) max0.I    = fabs(sky[i].I);
		if( fabs(sky[i].Q) > max0.Q    ) max0.Q    = fabs(sky[i].Q);
		if( fabs(sky[i].U) > max0.U    ) max0.U    = fabs(sky[i].U);
		if(  sky[i].Nobs   > max0.Nobs ) max0.Nobs = sky[i].Nobs;
	}
	cout << "\tmax0.I = " << max0.I << endl;
	cout << "\tmax0.Q = " << max0.Q << endl;
	cout << "\tmax0.U = " << max0.U << endl;
	cout << "\tmax0.N = " << max0.Nobs << endl;
	max = max0;

	char str[20];
	sprintf(str,"%8.2e",getmax());
	emit tellValueMax(str);

	selector.clear();
	setTexture();
	updateGL();

	return;
}
//-------------------------------------------------
void SkyViewer::reloadFile()
{
	cout << "\nSkyViewer::reloadFile\t" << currentfname << endl;

	if( ! sky.loadfitsfile(currentfname) ) {
		loadFile();
		return;
	}
	setTexture();
	updateGL();
	return;
}
//-------------------------------------------------
void SkyViewer::setDisplayMode(const QString & mode)
{
    cout << "SkyViewer::setDisplayMode:  " << mode << endl;
	if( mode == "3D Sphere" ) {
		current_mode = sphere;
		constraint->setRotationConstraintType(AxisPlaneConstraint::FREE);
		camera()->frame()->setConstraint(constraint);
		projection = &sphere3d_proj;
	}
	else if( mode == "Mollweide" ) {
		current_mode = mollweide;
		// fix rotation
		//camera()->setPosition(3.0,0,0);
		//camera()->setPosition( Vec(3.0,0,0) );
		//camera()->setOrientation(M_PI/2,0);
		//camera()->setUpVector(Vec(0,0,1));
		camera()->setPosition( Vec(3.0,0,0) );
		camera()->lookAt( Vec(0,0,0) );
		camera()->setUpVector(Vec(0,0,1));

		constraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN);
		camera()->frame()->setConstraint(constraint);
		projection = &mollweide_proj;
		selector.clear();
	}
	else
		return;
	setTexture();
	updateGL();
	return;
}
//-------------------------------------------------
void SkyViewer::setDisplayValue(const QString &value)
{
	DisplayValue old_value = current_value;
	if( value == "Temperature" )
		current_value = temperature;
	else if( value == "Q Polarization" ) {
		if( ! sky.has_polarization()) {
			cerr << "SkyViewer::setDisplayValue:No polarization data in loaded file." << endl;
			return;
		}
		current_value = Qpolarization;
	}
	else if( value == "U Polarization" ) {
		if( ! sky.has_polarization()) {
			cerr << "SkyViewer::setDisplayValue:No polarization data in loaded file." << endl;
			return;
		}
		current_value = Upolarization;
	}
	else if( value == "Num Obs" ) {
		if( ! sky.has_Nobs()) {
			cerr << "SkyViewer::setDisplayValue:No 'Num Obs' data in loaded file." << endl;
			return;
		}
		current_value = NumObs;
	}
	// Nothing to do
	if( current_value == old_value )
		return;
	setTexture();
	updateGL();
}
//-------------------------------------------------
void SkyViewer::setValueMax(int value)
{
	max.I = (0.01*value)*max0.I;
	max.Q = (0.01*value)*max0.Q;
	max.U = (0.01*value)*max0.U;
	max.Nobs = (0.01*value)*max0.Nobs;

	char str[20];
	sprintf(str,"%8.2e",getmax());
	emit tellValueMax(str);
	setTexture();
	updateGL();
	return;
}
//-------------------------------------------------
void SkyViewer::setValueMax(const QString & svalue)
{
	maxstr = svalue;
	cout << "svalue: " << svalue << endl;
	return;
}
//-------------------------------------------------
void SkyViewer::setValueMax2()
{
	cout << "SkyViewer::setValueMax2(): " << endl;
}
//-------------------------------------------------
void SkyViewer::setValueMax()
{
	cout << "SkyViewer::setValueMax(): " << maxstr << endl;
	bool ok;
	char str[20];
	float newvalue = maxstr.toFloat(&ok);
	if( ! ok ) {
		sprintf(str,"%8.2e",getmax());
		emit tellValueMax(str);
		return;
	}
	cout << "\t" << newvalue;
	float value = 0;
	switch( current_value ){
		case temperature: value = 100*newvalue/max0.I; break;
		case Qpolarization: value = 100*newvalue/max0.Q; break;
		case Upolarization: value = 100*newvalue/max0.U; break;
		case NumObs: value = 100*newvalue/max0.Nobs; break;
	}
	cout << "\t" << value << endl;
	emit tellValueMax((int)value);
	max.I = (0.01*value)*max0.I;
	max.Q = (0.01*value)*max0.Q;
	max.U = (0.01*value)*max0.U;
	max.Nobs = (0.01*value)*max0.Nobs;

	setTexture();
	updateGL();

	return;
}
//-------------------------------------------------
/*
void SkyViewer::setTextureRes(int v)
{
	tex_res = 2 << (6+v);
    //cout << "SkyViewer::setTextureRes:  " << v << ", res = " << tex_res << endl;
	setTexture();
	updateGL();
	return;
}
*/
//-------------------------------------------------
void SkyViewer::setSphereRes(int v)
{
	sphere_res = (2 << v);
	sphere3d_proj.setResolution(sphere_res);
	mollweide_proj.setResolution(sphere_res);	//sphere_res = v;
    cout << "SkyViewer::setSphereRes(int): " << v << "\t" << (2 << v) << endl;
	updateGL();
	return;
}
/*
//-------------------------------------------------
void SkyViewer::setSphereRes(const QString &strv)
{
	cout << "SkyViewer::setSphereRes: " << strv << endl;
	return;
}
*/
//-------------------------------------------------
void SkyViewer::showHelp()
{
	help();
}
/*
//-------------------------------------------------
void SkyViewer::tellValueMax(const QString & str){

}
*/
//====================================================================
// Service methods for managing textures
//====================================================================
void SkyViewer::setTexture()
{
	setSkyColor();
	texture.set(sky.nside(), sky.ordering(), skycolor);
	texture.highlight(selector.selectedPixel(),true);
	texture.glTexture();
	return;
}
//====================================================================
void SkyViewer::setSkyColor()
{
	if( skycolor.size() != sky.npixels() )
		skycolor.resize(sky.npixels());

	float v;
	int cidx = 0;
	uint blank = qRgb(0,0,0);
	//uint blank = qRgb(255,255,255);
	uint color;
	uint pix;
	//int ncolors = colors.size();
	for(pix = 0; pix < sky.npixels(); pix++) {
		if( sky.has_Nobs() && (sky[pix].Nobs < 1.0) ) {
			color = blank;
		}
		else {
			switch( current_value ){
				case temperature:
					v = sky[pix].I;
					cidx = (int)(0.5*(v/max.I+1)*ncolors);
					break;
				case Qpolarization:
					v = sky[pix].Q;
					cidx = (int)(0.5*(v/max.Q+1)*ncolors);
					break;
				case Upolarization:
					v = sky[pix].U;
					cidx = (int)(0.5*(v/max.U+1)*ncolors);
					break;
				case NumObs:
					v = sky[pix].Nobs;
					cidx = (int)(v/max.Nobs*ncolors);
					break;
			}
			if( cidx < 0 )
				cidx = 0;
			else if( cidx >= ncolors )
				cidx = ncolors - 1;
			color = colors[cidx];
		}
		skycolor[pix] = color;
	}
	return;
}
//====================================================================
void SkyViewer::postSelection(const QPoint& point)
{
	if( ! selector.select(point,camera(),projection) ) {
		texture.highlight(selector.selectedPixel(),false);
		texture.glTexture();
		emit tellPixelInfo(QString(""));
		return;
	}
	texture.highlight(selector.selectedPixel(),false);
	long pix = selector.toHealpix(sky);
	texture.highlight(pix,true);
	texture.glTexture();

	QString text = QString("Pixel: %1 (%2)")
		.arg(pix)
		.arg(sky.ordering() ? "ring" : "nest" );
	double theta,phi;
	sky.pix2ang(pix, &theta, &phi);
	text += QString("\nTheta: %1 (deg)\nPhi: %2 (deg)")
			.arg((180/M_PI)*theta).arg((180/M_PI)*phi);

	if( sky.has_polarization() ) {
		text += QString("\nI: %1\nQ: %2\nU: %3")
			.arg(sky[pix].I).arg(sky[pix].Q).arg(sky[pix].U);
	}
	else {
		text += QString("\nT: %1").arg(sky[pix].I);
	}
	if( sky.has_Nobs() )
		text += QString("\nNobs: %1").arg(sky[pix].Nobs);

	emit tellPixelInfo(text);
	return;
}
